	list	P=16F84, N=100

; *******************************************************************************************
; Program:	HI_TACHO.ASM
; Author:	John Baraclough
; Last update:	2000-02-13
;
; This program is for a PIC16F84 and is intended to measure the rotation of an eight blade
; sensor running at up to 7650rpm. Although it will measure down to 1rpm, accuracy at the low
; end is limited.
;
; The tacho blade input is on RB7 and uses the PortB change interrupt. There are sixteen
; interrupts for every revolution of the shaft and these are counted. This count is sampled
; and reset every 1/8 of a second. Thirty such samples are stored in a circular buffer and
; each time a new one is written the whole buffer is summed to give a direct reading of RPM.
; This value is converted to BCD and displayed on a 4-digit 7-segment LED display with a
; serial interface. There is no interaction with this display so it is optional and will not
; affect the operation if it is not fitted.
;
; The timing is derived from a 7.3278MHz external crystal oscillator. This gives an
; instruction clock of 1.8432MHz with a cycle time of 542.5nSec. The instruction clock is
; divided by 1024 (256*4) in TMR0 to generate an interrupt every 1/1800 of a second. These
; are counted and when 225 such interrupts have been received 1/8 of a second has passed.
;
; Communication with the host is via an half-duplex RS485 link using a MAX3100 UART. The UART
; is driven with a 3.6846MHz clock derived from the PIC clock. The UART interrupt output is
; connected to RB0 and uses the PIC external INT. Communication with the UART is via a
; software emulated SPI.
;
; NOTE:- This version uses polling for the UART receiver and not interrupts. The 'LineIdle'
; count may need to be higher for this mode than if interrupts are used.
; *******************************************************************************************

	include	<p16f84.inc>
	Radix dec
	errorlevel 2

	__CONFIG _CP_OFF & _WDT_OFF & _XT_OSC
	

; *******************************************************************************************
; Vectors and other preset variables.
; *******************************************************************************************
ResetVect	equ	0x0000
IntVect		equ	0x0004

OptionData	equ	0x81			; TMR0 prescalar = 4,  PORTB pull-ups enabled.

PortaData	equ	0x08			; UART disabled and all other bits low.
TrisaData	equ	0x02			; RA1 => IN, RA4..2 & RA0 => OUT.
PortbData	equ	0x40			; LED display disabled and all other bits low.
TrisbData	equ	0x81			; RB7 & RB0 => IN, RB6..1 => OUT.

; *******************************************************************************************
; UART interrupt is not used yet. jnb 2000-01-31.
; *******************************************************************************************
;IntControl	equ	(1<<GIE)+(1<<T0IE)+(1<<INTE)+(1<<RBIE)	; Interrupt control register
IntControl	equ	(1<<GIE)+(1<<T0IE)+(1<<RBIE)	; Interrupt control register

;UartControl	equ	0xc40b			; !FEN + !SHDN + !TM + RM + !PM + !RAM
						; !IR  + !ST + !PE + !L + B3 + !B2 + B1 + B0
UartControl	equ	0xc00b			; !FEN + !SHDN + !TM + !RM + !PM + !RAM
						; !IR  + !ST + !PE + !L + B3 + !B2 + B1 + B0

UartRdConfig	equ	0x4000			; Read configuration command.
UartRtsPinLo	equ	0x8600			; Disable Tx and set RTS pin low.
UartRtsPinHi	equ	0x8400			; Disable Tx and set RTS pin high.
UartWrDat	equ	0x8000			; Enable Tx and set RTS pin high.
UartRdDat	equ	0x0000			; Read data command.

IdleTime	equ	10			; This is the number of interrupts timer counted
						; before the 'LII' flag is set.

OscFreq		equ	7372800			; Oscillator crystal needed for MAX3100.

D1Char		equ	3			; Delay loop count for 1 character.

T0Count		equ	225			; This is the number of interrupts for 1/8 sec.

BuffSize	equ	30			; Thirty samples in the buffer.

DecPoint	equ	0x80


; *******************************************************************************************
; Bit definitions for PORTA
; *******************************************************************************************
UartEna		equ	3
UartClock	equ	2
UartRx		equ	1			; This is data FROM Uart TO PIC
UartTx		equ	0			; This is data FROM PIC TO Uart


; *******************************************************************************************
; Bit definitions for UartStat after 'RdConfig' has been called.
; *******************************************************************************************
RxRdy		equ	1			; This is data FROM Uart TO PIC
TxRdy		equ	0			; This is data FROM PIC TO Uart


; *******************************************************************************************
; Bit definitions for PORTB
; *******************************************************************************************
LedEna		equ	6
LedData		equ	5
LedClock	equ	4


; *******************************************************************************************
; Bit definitions for count flags register
; *******************************************************************************************
BCO		equ	7			; 'BladeCountOverflow' in current count.
NDR		equ	6			; 'NewDataReady' from interrupt.
SDS		equ	5			; 'ShowDisplayString' task flag.
BUS		equ	4			; 'BuildUartString' task flag.
OFE		equ	3			; 'OverFlowError' flag.
MIP		equ	2			; 'MessageInProgress' flag.
LII		equ	1			; 'LineIsIdle' flag.
LZS		equ	0			; 'LeadingZeroSuppression' flag.

; *******************************************************************************************
; Data storage block
;
; Note:- 16 bit words are stored as HiByte, LoByte in true 'Motorola hi-endian' fashion.
; *******************************************************************************************
	cblock	0x0c

		Flags				; See bit definitions above.
		PortbCopy			; Shadow register for PORTB.
		Timer				; 8 bit TMR0 overflow counter.
		Count				; These two are used in the Revs to ...
		Temp				; ... BcdRevs conversion and display routines.
		Blade				; Counter location for blade interrupt.
		BuffPtr				; Pointer to the offset of the current buffer.
		Buffer: BuffSize		; Circular buffer.
		Revs: 2				; 16 bit total count for the 3 3/4 ...
						; ... seconds of blade counts (==RPM).
		BcdRevs: 2			; BCD conversion of 'Revs'.
		Rats: 4				; Stores the overflow of the last 32 samples.
		SaveW				; Storage area for the registers that ...
		SaveStat			; ... are destroyed in the interrupt ...
		SaveFSR				; ... service routine.
		UartComm: 2			; 16 bit buffer for commands to MAX3100.
		UartStat: 2			; 16 bit buffer for status from MAX3100.
		LineIdle			; Timer counter for RS485 idle line.
		UartPtr				; Pointer to current location in 'UartBuff'.
		UartBuff: 13			; RS485 message receive buffer.
		Checksum			; Temporary store for the checksum.

	endc

; *******************************************************************************************
; Local macros which are only used in this file.
; *******************************************************************************************

Bank0		macro
			bcf	STATUS, RP0
		endm

Bank1		macro
			bsf	STATUS, RP0
		endm

EnableDisplay	macro
			bcf	PortbCopy, LedEna
			movf	PortbCopy, W
			movwf	PORTB
		endm

DisableDisplay	macro
			bsf	PortbCopy, LedEna
			movf	PortbCopy, W
			movwf	PORTB
		endm

SetStartBit	macro
			bsf	PortbCopy, LedData
			movf	PortbCopy, W
			movwf	PORTB
		endm

StrobeDisplay	macro
			movf	PortbCopy, W
			iorlw	(1<<LedClock)
			movwf	PORTB
			andlw	~(1<<LedClock)
			movwf	PORTB
		endm

EnableUart	macro
			bcf	PORTA, UartEna
		endm

DisableUart	macro
			bsf	PORTA, UartEna
		endm

StrobeUartHi	macro
			bsf	PORTA, UartClock
		endm

StrobeUartLo	macro
			bcf	PORTA, UartClock
		endm


; *******************************************************************************************
; Program start.
; *******************************************************************************************
	org	ResetVect

		goto	Main

		call	0x00			; Year is 2000!
		movf	0x02, W			; Month
		movf	0x11, W			; Day

; *******************************************************************************************
; Interrupt service routine.
;
; Arrive here when one of the following happens.
;
; 1. PORTB bit 7 changes state.
;	This is a blade interrupt and will happen on both edges of the blade.
;
; 2. TMR0 overflows.
;	This will happen at 1800Hz or 225 times every 1/8 second. After 225 interrupts the
;	current blade count is transferred to the next location in the circular buffer, the
;	blade count is cleared, the timer overflow counter is reloaded and the 'NDR' flag is
;	set to indicate that the buffer has been updated.
;
; 3. PORTB bit 0 goes low.
;	This is the UART interrupt.
; *******************************************************************************************
IntServ		movwf	SaveW			; Save ALL the registers that are affected.
		movf	STATUS, W		; This changes the status register but ...
		movwf	SaveStat		; ... since it is already saved there is ...
		movf	FSR, W			; ... no problem.
		movwf	SaveFSR
		

; *******************************************************************************************
; Blade interrupt handler.
; *******************************************************************************************
RBChangeInt	btfss	INTCON, RBIF		; Is it a blade interrupt?
		goto	TimerInt		; If not jump on to timer.
		movf	PORTB, W		; Otherwise read the port to reset the ...
		bcf	INTCON, RBIF		; ... source and clear the interrupt flag.
		incf	Blade, F		; Bump the blade count, if it overflows ...
		btfsc	STATUS, Z		; ... to zero then ...
		bsf	Flags, BCO		; ... set the 'BladeCountOverflow' flag ...
		goto	IntEnd			; ... then exit.


; *******************************************************************************************
; Timer interrupt handler.
; *******************************************************************************************
TimerInt	btfss	INTCON, T0IF		; Is it a TMR0 interrupt?
		goto	SerialInt		; If not it must be the UART.
		bcf	INTCON, T0IF		; Clear the interrupt flag.

		btfss	Flags, MIP		; Is 'MessageInProgress' flag set?
		goto	TimerInt1		; If not do nothing here.
		decfsz	LineIdle, F		; Decrement 'LineIdle' counter and if not ...
		goto	TimerInt1		; ... zero yet then continue with timer.
		bcf	Flags, MIP		; Clear 'MessageInProgress' flag and ...
		bsf	Flags, LII		; ... set 'LineIsIdle' flag.

TimerInt1	decfsz	Timer, F		; Decrement interrupt counter and exit if ...
		goto	IntEnd			; ... not zero.

TimeOut		movlw	T0Count			; Reload TMR0 interrupt counter.
		movwf	Timer
		movlw	Buffer - 1		; Calculate the buffer pointer value for ...
		addwf	BuffPtr, W		; ... the location of the most recent ...
		movwf	FSR			; ... blade count.
		movf	Blade, W		; Get the current blade count, ...
		movwf	INDF			; ... store it in the buffer and ...
		clrf	Blade			; ... clear the blade count.

		bsf	Flags, NDR		; Set the 'NewDataReady' flag.
		decfsz	BuffPtr, F		; Decrement the pointer and if not zero ...
		goto	IntEnd			; ... then exit.
		movlw	BuffSize		; Otherwise reload the pointer and ...
		movwf	BuffPtr
		goto	IntEnd			; ... exit.


; *******************************************************************************************
; UART interrupt handler - This code is not used and commented out. jnb 2000-01-31.
; *******************************************************************************************
SerialInt	;bcf	INTCON, INTF		; Clear the interrupt flag.
;		bsf	Flags, MIP		; Set 'MessageInProgress' flag.
;		call	ReadAscii		; Get character from UART.


; *******************************************************************************************
; End of interrupt.
; *******************************************************************************************
IntEnd		movf	SaveFSR, W		; Restore the FSR register, ...
		movwf	FSR
		movf	SaveStat, W		; ... the STATUS register and finally ...
		movwf	STATUS
		swapf	SaveW, F		; ... the W register without affecting ...
		swapf	SaveW, W		; ... the STATUS register flags.
		retfie				; Then return from the interrupt.


; *******************************************************************************************
; Binary to BCD conversion routine.
;
;	This is the 'famous' add three and shift routine. This routine has no useful comments
;	because I don't know how it works.
;
; Input: 'Revs' contains the current RPM value in 16 bit binary form.
;
; Output: 'BCDRevs' contains the 4 digit BCD version.
;
; Registers: FSR, Temp & Count
;
; Flags: None
;
; Stack usage: 3 levels
; *******************************************************************************************
Bin2Bcd16	bcf	STATUS, C		; Clear the carry bit.
		movlw	16			; Sixteen bits to convert.
		movwf	Count
		clrf	BcdRevs+1
		clrf	BcdRevs

BcdLoop		rlf	Revs + 1, F
		rlf	Revs, F
		rlf	BcdRevs + 1, F
		rlf	BcdRevs, F

		decfsz	Count, F		; Done sixteen bits yet?
		goto	AdjDec			; No so adjust BCD.
		return				; Otherwise finish.

AdjDec		movlw	BcdRevs + 1		; Do low byte first.
		movwf	FSR
		call	AdjBcd
		decf	FSR, F			; Now do high byte.
		call	AdjBcd
		goto	BcdLoop

AdjBcd		movf	INDF, W			; Get BCD number.
		addlw	0x03			; Add 3 to low BCD digit.
		movwf	Temp
		btfsc	Temp, 3			; Test if low nibble is > 7.
		movwf	INDF			; Overwrite value if it is.
		movf	INDF,W
		addlw	0x30			; Add 3 to high BCD digit.
		movwf	Temp
		btfsc	Temp, 7			; Test if high nibble > 7.
		movwf	INDF			; Overwrite value if it is.
		return


; *******************************************************************************************
; Character display routine.
;
;	This is for displaying one character on a four digit seven segment LED with a serial
;	interface, such as the III-V TMS5374, or any other display using the NatSemi MM5450
;	chip.
;
; Input:	Disp1Numb: 'W' contains the number in the lowest nibble.
;		Disp1Char: 'W' contains the bit pattern of the character.
;
; Output: None.
;
; Registers: Count & Temp.
;
; Flags: None.
;
; Stack usage: None.
; *******************************************************************************************
Disp1Numb	andlw	0x0f			; Mask off least significant bits and ...
		btfss	STATUS, Z		; Is it zero?
		goto	DispNotZero		; No, so continue.
		btfsc	Flags, LZS		; Yes, so test 'LeadingZeroSuppression' flag.
		goto	Disp1Char		; If flag is set then bit pattern is correct.

DispNotZero	bcf	Flags, LZS		; Clear 'LeadingZeroSuppression' flag.
		call	ConvertChar		; ... get bit pattern.

Disp1Char	movwf	Temp			; Save bit pattern.
		movlw	8			; Eight bits to send.
		movwf	Count

Disp1Loop	bcf	PortbCopy, LedData	; First clear data bit ...
		rrf	Temp, F			; Now get data bit and ...
		btfsc	STATUS, C		; ... test if it is high.
		bsf	PortbCopy, LedData	; Set display data bit if it is.
		movf	PortbCopy, W
		movwf	PORTB
		StrobeDisplay
		decfsz	Count, F		; Done eight bits yet?
		goto	Disp1Loop		; No, so go round again.
		return


; *******************************************************************************************
; Delay routine.
;
;	This is for delaying by N lots of 424uS (assuming 7.3728MHz clock). This routine is
;	called after a message has been sent from the UART. It is neccessary as there is no
;	flag in the MAX3100 to show that the transmitter register is empty.
;
; Input: None.
;
; Output: None.
;
; Registers: Temp & Count.
;
; Flags: None.
;
; Stack usage: None.
; *******************************************************************************************
Delay1Char	movlw	D1Char

DelayLoop	decfsz	Temp, F
		goto	DelayLoop
		decfsz	Count, F
		goto	DelayLoop
		return


; *******************************************************************************************
; UART interface routine.
;
;	This is for sending a command to and retrieving the result from the MAX3100 UART.
;
;	There are five entry points which are called.
;
;	2. SendAscii: 	sends the character in W.
;
;	3. ReadAscii: 	reads a character from the UART buffer. The character is returned
;			in the UartStat+1 register.
;
;	4. ReadStat: 	gets the Rxr & Txr status bits.
;
;	5. UartLink16:	is the lowest level entry used for initialisation.
;
; Input: UartComm contains the command and UartComm+1 the data for the UART.
;
; Output: UartStat contains the status and UartStat+1 the data from the UART.
;
; Registers: Count, Temp, UartComm, UartComm+1, UartStat & UartStat+1.
;
; Flags: None.
;
; Stack usage: 2 levels.
; *******************************************************************************************
TxReady		call	RdConfig		; Wait for 'T' bit to be set
		btfss	UartStat+1, TxRdy
		goto	TxReady
		return

SetRs485Tx	movlw	high (UartRtsPinHi)	; No need to wait here.
		goto	RdAsc1

SetRs485Rx	call	TxReady			; Wait until Tx buffer is empty.
		call	Delay1Char		; Wait until Tx register is empty.
SetRx1		movlw	high (UartRtsPinLo)
		goto	RdAsc1

SendAscii	movwf	Temp			; ... save it.
		call	TxReady
		movf	Temp, W			; Recover ASCII character and ...
		movwf	UartComm+1		; ... put it in the command buffer.

WriteData	movlw	high (UartWrDat)	; Set the write data command.
		movwf	UartComm
		goto	UartLink16

ReadAscii	movlw	high (UartRdDat)	; Set the read data command.
RdAsc1		movwf	UartComm
		clrf	UartComm+1		; Clear the low byte.

UartLink16	movlw	16			; Sixteen bits to send.
		movwf	Count
		goto	UartStart

RdConfig	movlw	high (UartRdConfig)	; Read configuration command high byte.
		movwf	UartComm
		movlw	2			; Two bits to send.
		movwf	Count

UartStart	EnableUart

UartLoop	bcf	PORTA, UartTx		; Clear data to UART.
		rlf	UartComm+1, F		; Rotate low byte then rotate ...
		rlf	UartComm, F		; ... high byte to get MSB to carry.
		btfsc	STATUS, C		; Is carry bit set?
		bsf	PORTA, UartTx		; If so then set data bit high.
		StrobeUartHi
		bcf	STATUS, C		; Clear carry.
		btfsc	PORTA, UartRx		; Test input data from UART and ...
		bsf	STATUS, C		; ... if high then set carry.
		rlf	UartStat+1, F		; Move carry bit into Uart status ...
		rlf	UartStat, F		; ... file register.
		StrobeUartLo
		decfsz	Count, F		; Done 16 bits yet?
		goto	UartLoop		; No, so go round again.
		DisableUart			; Yes, so disable UART and ...
		return				; ... exit.


; *******************************************************************************************
; Binary to ASCII hexadecimal conversion routine.
;
;	This is for converting the low nibble of the W register to ASCII hex. The result is
;	stuffed into a buffer. If the 'LeadingZeroSuppression' flag is set the digit is
;	skipped. A non-zero digit will clear the flag so that subsequent zeros will be sent
;	correctly. The routine should be entered with the 'LeadingZeroSuppression' flag
;	cleared when sending the final digit of a sequence, just in case it is a zero.
;
; Input: W contains the data, FSR is pointing to the buffer location.
;
; Output: W contains the ASCII character, which has been put into the buffer, and the pointer
;		has been incremented.
;
; Registers: Temp.
;
; Flags: None.
;
; Stack usage: None.
; *******************************************************************************************
Bin2Hex		andlw	0x0f			; Mask off the lower nibble.
		movwf	Temp			; Save it.
		btfss	STATUS, Z		; Is it zero?
		goto	BinNotZero		; No, so continue.
		btfss	Flags, LZS		; Yes, so test 'LeadingZeroSuppression' flag.
		goto	NoLzsFlag		; If flag not set then continue, ...
		return				; ... otherwise do nothing and exit.

BinNotZero	bcf	Flags, LZS		; Clear 'LeadingZeroSuppression' flag.
NoLzsFlag	movlw	0x0a			; Is number greater than 9?
		subwf	Temp, W			; Carry will be set if it is.
		movlw	0x30			; Offset for 0->9.
		btfsc	STATUS, C		; Is it hex A->F?
		addlw	0x07			; Yes so adjust offset.
		addwf	Temp,W			; Make ASCII digit.
		movwf	INDF			; Put it in the buffer and ...
		incf	FSR, F			; ... bump the pointer.
		addwf	Checksum, F		; Adjust the checksum.
		return


; *******************************************************************************************
; ASCII hexadecimal to binary conversion routine.
;
;	This is for converting an ASCII hex digit to binary in the low nibble of W.;
;
; Input: W contains the ASCII character.
;
; Output: W contains the binary data. The carry flag is set if the input character
;		is not in the range '0'->'9' or 'A'->'F'.
;
; Registers: Temp.
;
; Flags: None.
;
; Stack usage: None.
; *******************************************************************************************
Hex2Bin		addlw	0xd0		; Normalise to '0' == 0.
		movwf	Temp
		btfss	STATUS, C	; If carry set then > '/'.
		goto	BadHex

		movlw	0x0a		; Now test for '0' >= X <= '9'.
		subwf	Temp, W		; If good data carry is clear.
		btfss	STATUS, C
		goto	GoodHex

		movlw	0xf9		; Now adjust for 'A'-> 'F'.
		addwf	Temp, F

		movlw	0x0a		; Test for < 'A'.
		subwf	Temp, W		; Carry set if good data.
		btfss	STATUS, C
		goto	BadHex

		movlw	0x10		; Test for > 'F'.
		subwf	Temp, W		; Carry set if good data.
		btfss	STATUS, C
		goto	GoodHex

BadHex		bsf	STATUS, C	; Error so set carry.
		return

GoodHex		movf	Temp, W		; Recover data.
		bcf	STATUS, C	; No error so clear carry.
		return


; *******************************************************************************************
; Main program starts here.
;
;	First initialise the variables and ports and then sit for forever in a loop doing:-
;
; 1. Check NewDataReady flag and if set do:-
;	a. Sum the contents of the 30 byte buffer into Revs (16 bit).
;	b. Convert Revs to 4 digit BCD in BcdRev.
;	c. Show the new value of BcdRevs on the LED display.
;	d. Go back to check flags again.
;
; 2.
;
; 3.
;
; 4.
;
; *******************************************************************************************
Main		movlw	PortaData
		movwf	PORTA
		movlw	PortbData
		movwf	PortbCopy		; Initialise copy of Port B.
		movwf	PORTB
		movlw	high DigitData		; Set up PCLATH for message display.
		movwf	PCLATH

		Bank1
		movlw	TrisaData
		movwf	TRISA
		movlw	TrisbData
		movwf	TRISB
		movlw	OptionData
		movwf	OPTION_REG
		Bank0

InitUart	call	SetRx1			; Set RS485 buffer to receive.
		movlw	high UartControl	; Initialise the UART.
		movwf	UartComm		; High byte first ...
		movlw	low UartControl		; ... then low byte.
		movwf	UartComm+1
		call	UartLink16		; Now send it.

InitRam		clrf	Flags		; Clear all the flags, ...
		clrf	Flags
		clrf	Rats			; ... the overflow flags, ...
		clrf	Rats+1
		clrf	Rats+2
		clrf	Rats+3
		clrf	Blade			; ... the current blade count and ...

		movlw	BuffSize		; ... the blade count buffer.
		movwf	Count
		movlw	Buffer
		movwf	FSR

ClearLoop	clrf	INDF
		incf	FSR, F
		decfsz	Count, F
		goto	ClearLoop

		movlw	UartBuff		; Set the UART pointer for the first message.
		movwf	UartPtr

SetupIsr	movlw	BuffSize		; Initialise the pointer for the ISR.
		movwf	BuffPtr
		movlw	T0Count			; Set up loop counter for timer interrupts.
		movwf	Timer
		movf	PORTB, W		; Read PortB to clear any change interrupt.
		movlw	IntControl
		movwf	INTCON			; Enable all the interrupts


; *******************************************************************************************
; Main program loop is here.
; *******************************************************************************************
MainLoop	call	RdConfig		; Check if there is a character in the UART.
		btfsc	UartStat+1, RxRdy	; If 'RxRdy' is set there is a character ...
		goto	GetChar			; ... waiting so get it.

		btfsc	Flags, LII		; Is a new message ready?
		goto	GotMessage		; Yes, so jump.

		btfsc	Flags, NDR		; Is NewDataReady set?
		goto	CalcRpm			; Yes, so jump.

		btfsc	Flags, SDS		; Is ShowDisplayString set?
		goto	ShowDisplay		; Yes, so jump.

		goto	MainLoop		; Nothing to do this time so go round again.


; *******************************************************************************************
; Got a new character in the UART so move it into the receive buffer.
; *******************************************************************************************
GetChar		call	ReadAscii		; Now read the next character.
		movf	UartPtr, W		; Get the pointer, ...
		movwf	FSR			; ... put it in FSR and ...
		incf	UartPtr, F		; ... bump the value.
		movf	UartStat+1, W		; Now get the character ...
		movwf	INDF			; ... and store it.
		movlw	IdleTime		; Set up line idle counter and ...
		movwf	LineIdle
		bsf	Flags, MIP		; ... set 'MessageInProgress' flag.
		goto	MainLoop		; ... go round again.


; *******************************************************************************************
; Got a new message in the receive buffer so deal with it.
; *******************************************************************************************
GotMessage	bcf	Flags, LII		; Clear 'LineIsIdle' flag.
		movlw	UartBuff		; Get the address of the start of the buffer ...
		movwf	FSR			; ... and initialise the pointer.

		movlw	'T'			; Is the message for the Tacho?
		xorwf	INDF, W
		btfss	STATUS, Z		; 'Z' flag will be set if it is.
		goto	ExitMessage		; If not then message is not for us.

TachoMessage	decf	UartPtr, F		; Adjust UART buffer pointer to point to ...
		decf	UartPtr, F		; ... the checksum.
		movlw	UartBuff		; Get the address of the start of the buffer ...
		subwf	UartPtr, W		; ... subtract the current buffer pointer to get ...
		movwf	Count			; ... the number of chars in the buffer.

		movf	UartPtr, W		; Initialise pointer.
		movwf	FSR

GetChecksum	movf	INDF, W			; Get high nibble of checksum and ...
		call	Hex2Bin			; ... convert it to binary.
		btfsc	STATUS, C		; The carry is set for a bad hex ...
		goto	SendAnswer9		; ... character, so send error 9.
		movwf	Checksum		; If OK put it in the lower nibble and ...
		swapf	Checksum, F		; ... swap it to high nibble.
		incf	FSR, F
		movf	INDF, W			; Now get the lower nibble and ...
		call	Hex2Bin			; ... convert it to binary.
		btfsc	STATUS, C		; The carry is set for a bad hex ...
		goto	SendAnswer9		; ... character, so send error 9.
		iorwf	Checksum, F		; If OK combine the nibbles.

		movlw	UartBuff		; Get the address of the start of the buffer ...
		movwf	FSR			; ... and initialise the pointer.

CheckLoop	movf	INDF, W			; Get next character of message, ...
		addwf	Checksum, F		; ... add to checksum, ...
		incf	FSR, F			; ... bump pointer, ...
		decfsz	Count, F		; ... decrement counter and ...
		goto	CheckLoop		; ... go round again if not finished.

		movf	Checksum, F		; If checksum is zero then the ...
		btfss	STATUS, Z		; ... 'Z' flag is set so we are OK.
		goto	SendAnswer9		; Bad checksum so send error 9.

TestCommand	movlw	UartBuff+1		; Get the address of the command in the buffer ...
		movwf	FSR			; ... and initialise the pointer.

		movlw	'0'			; Is it command '0'?
		xorwf	INDF, W
		btfsc	STATUS, Z		; 'Z' flag is set if it is ...
		goto	SendAnswer0		; ... so go and do it.

		movlw	'1'			; Otherwise see if it is command '1'.
		xorwf	INDF, W
		btfss	STATUS, Z		; 'Z' flag is set if it is ...
		goto	SendAnswer9		; ... but if not it is a bad command.

SendRpm		btfsc	Flags, OFE		; If 'OverFlowError' flag is set ...
		goto	SendAnswer8		; ... send 'System Error 1'.

		movlw	UartBuff		; Otherwise data is OK, so ...
		movwf	FSR			; ... initialise pointer, ...
		movlw	't'			; ... initialise message and ...
		movwf	Checksum		; ... initialise the checksum.
		movwf	INDF
		incf	FSR, F
		movlw	'1'
		addwf	Checksum, F		; Adjust the checksum.
		movwf	INDF
		incf	FSR, F
		bsf	Flags, LZS		; Set 'LeadingZeroSuppression' flag.
		swapf	BcdRevs, W		; Get most significant digit and ...
		call	Bin2Hex			; ... move ASCII character to buffer.
		movf	BcdRevs, W
		call	Bin2Hex
		swapf	BcdRevs+1, W
		call	Bin2Hex
		bcf	Flags, LZS		; Clear 'LeadingZeroSuppression' flag.
		movf	BcdRevs+1, W		; Get least significant digit and ...
		call	Bin2Hex			; ... move ASCII character to buffer.

AddChecksum	comf	Checksum, F		; Invert checksum.
		incf	Checksum, F
		swapf	Checksum, W		; Get most significant digit of checksum and ...
		call	Bin2Hex			; ... move ASCII character to buffer.
		movf	Checksum, W		; Get least significant digit of checksum and ...
		call	Bin2Hex			; ... move ASCII character to buffer.
		clrf	INDF
		goto	UartMessage

SendAnswer0	movlw	Message0-MessageData	; Pointer to message.
		goto	MoveMessage

SendAnswer8	movlw	Message8-MessageData	; Pointer to message.
		goto	MoveMessage

SendAnswer9	movlw	Message9-MessageData	; Pointer to message.

MoveMessage	movwf	Temp
		movlw	UartBuff
		movwf	FSR

MoveMessLoop	movf	Temp, W			; Get pointer.
		call	GetMessageData		; Get data.
		movwf	INDF			; Put it in the buffer.
		incf	Temp, F
		incf	FSR, F
		andlw	0xff
		btfss	STATUS, Z
		goto	MoveMessLoop

UartMessage	call	SetRs485Tx		; Set RS485 transmitter on.
		movlw	UartBuff		; Initialise pointer.
		movwf	FSR

UartMessLoop	movf	INDF, W			; Get next message character.
		andlw	0xff			; Test for end of message ...
		btfsc	STATUS, Z		; ... when Z flag is set ...
		goto	UartMessEnd		; ... it is the end of the message.
		call	SendAscii		; Otherwise send character to UART, ...
		incf	FSR, F			; ... bump pointer and ...
		goto	UartMessLoop		; ... go round again.

UartMessEnd	call	SetRs485Rx

ExitMessage	movlw	UartBuff		; Reset pointer for next message.
		movwf	UartPtr
		goto	MainLoop		; Nothing to do here yet.


; *******************************************************************************************
; Got a new count in  the buffer so it is time to calculate the new RPM value.
; *******************************************************************************************
CalcRpm		bcf	Flags, NDR		; Clear NewDataReady flag, which is ...
		movlw	BuffSize		; ... how we got here. Then set up ...
		movwf	Count			; ... to sum the whole buffer.
		movlw	Buffer
		movwf	FSR
		clrw				; Clear W register and ...
		movwf	Revs			; ... high byte of result.

AddLoop		addwf	INDF, W			; Add next value
		btfsc	STATUS, C		; Test for overflow and increment ...
		incf	Revs, F			; ... high byte if one occurred.
		incf	FSR, F			; Bump pointer.
		decfsz	Count, F		; Have we done the whole buffer yet?
		goto	AddLoop			; No, so go round again.
		movwf	Revs+1			; Yes, finished so store low byte.
		call	Bin2Bcd16		; Now convert to BCD.
		rlf	Flags, W		; Get the 'BladeCountOverflow' flag, leaving ...
		rlf	Rats+3, F		; ... the 'Flags' register unchanged. Then ...
		rlf	Rats+2, F		; ... rotate it into the 32 bit error buffer.
		rlf	Rats+1, F
		rlf	Rats, F
		bcf	Flags, BCO		; Clear BCO flag for next blade count.
		bcf	Flags, OFE		; Clear 'OverFlowError' flag.
		movf	Rats, W			; Now check if there are any of ...
		iorwf	Rats+1, W		; ... the 32 bits set.
		iorwf	Rats+2, W
		iorwf	Rats+3, W
		btfss	STATUS, Z		; If Z flag set, all is OK, otherwise ...
		bsf	Flags, OFE		; Set 'OverFlowError' flag.

		bsf	Flags, SDS		; Set ShowDisplayString flag and ...
		goto	MainLoop		; ... go round again.


; *******************************************************************************************
; Time to show the new data on the LED display.
; *******************************************************************************************
ShowDisplay	bcf	Flags, SDS		; Clear ShowDisplayString flag.
		EnableDisplay
		SetStartBit
		StrobeDisplay
		btfsc	Flags, OFE		; Test 'OverFlowError' flag and if it is ...
		goto	ShowOverflow		; ... set there is bad data present.

ShowGoodData	bsf	Flags, LZS		; Set 'LeadingZeroSuppression' flag.
		swapf	BcdRevs, W		; Get most significant digit ...
		call	Disp1Numb		; ... and display it.
		movf	BcdRevs, W		; Get next digit ...
		call	Disp1Numb		; ... and display it.
		swapf	BcdRevs+1, W		; Get next digit ...
		call	Disp1Numb		; ... and display it.
		bcf	Flags, LZS		; Clear 'LeadingZeroSuppression' flag.
		movf	BcdRevs+1, W		; Get least significant digit ...
		call	Disp1Numb		; ... and display it.
		goto	DispEnd

ShowOverflow	movlw	DecPoint		; Decimal point
		call	Disp1Char		; ... display it.
		movlw	DecPoint		; Decimal point
		call	Disp1Char		; ... display it.
		movlw	DecPoint		; Decimal point
		call	Disp1Char		; ... display it.
		movlw	DecPoint		; Decimal point
		call	Disp1Char		; ... display it.

DispEnd		movlw	3			; Three extra strobes are needed to ...
		movwf	Count			; ... complete the transfer.
		bcf	PORTB, LedData		; Send zeros for last three bits.

DispEndLoop	StrobeDisplay
		decfsz	Count, F		; Finished yet?
		goto	DispEndLoop		; No so go round again.
		DisableDisplay			; Yes so disable display and ...
		goto	MainLoop		; ... go round again.


; *******************************************************************************************
; Display look-up table.
;
;	These are the bit patterns for displaying digits on a four digit seven
;	segment LED with a serial interface, such as the III-V TMS5374, or any
;	other display using the NatSemi MM5450 chip.
; *******************************************************************************************
		org	0x037f

ConvertChar	addwf	PCL, F			; Calculate the pointer.
DigitData	retlw	0x3f			; 0 = _fedcba
		retlw	0x06			; 1 = ____cb_
		retlw	0x5b			; 2 = g_ed_ba
		retlw	0x4f			; 3 = g__dcba
		retlw	0x66			; 4 = gf__cb_
		retlw	0x6d			; 5 = gf_dc_a
		retlw	0x7d			; 6 = gfedc_a
		retlw	0x07			; 7 = ____cba
		retlw	0x7f			; 8 = gfedcba
		retlw	0x6f			; 9 = gf_dcba
		retlw	0x5f			; a = g_edcba
		retlw	0x7c			; b = gfedc__
		retlw	0x58			; c = g_ed___
		retlw	0x5e			; d = g_edcb_
		retlw	0x7b			; e = gfed_ba
		retlw	0x71			; f = gfe___a

; *******************************************************************************************
; Message look-up table.
; *******************************************************************************************
GetMessageData	addwf	PCL, F
MessageData
Message0	retlw	't'
		retlw	'0'
		retlw	'5'
		retlw	'C'
		retlw	0x00

Message8	retlw	't'
		retlw	'8'
		retlw	'0'
		retlw	'1'
		retlw	'F'
		retlw	'3'
		retlw	0x00

Message9	retlw	't'
		retlw	'9'
		retlw	'5'
		retlw	'3'
		retlw	0x00

	end

